#line 2 "guess.c"
/*                                                                      
 *  C-SaCzech
 *
 *  Copyright (c) 1996-2000 Jaromir Dolecek <dolecek@ics.muni.cz>           
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ----
 * Note:
 *   If the conditions of GPL are unacceptable for you, contact
 *   author to adjust something more suitable for your needs.
 */

/* $Id: guess.c,v 1.4 2000/01/24 13:18:03 dolecek Exp $ */

#include "csacek.h"

/*
 * tries to guess encoding and language supported by remote client; uses
 * ``Accept-Charset'' and ``Accept-Charset'' headers, if client has sent them
 * if it's not the case, ISO-8859-2 is forced for people from within
 * .cz and ASCII for others
 */
void
csa_toguess(p, redirect_client)
  csa_params_t *p;
  int redirect_client;
{
	const char *charset=NULL, *suffix=NULL, *tocut, *prefix;
	char *newurl;
	const csa_String *strp;
	cstools_t outenc=CSTOOLS_UNKNOWN;

#ifdef CSA_DEBUG
	csa_debug(p->dbg, "csa_toguess: called");
#endif

	/* set indication charset & language was guessed */
	CSA_SET(p->flags, CSA_FL_GUESSED_CHARSET);

	/* test language preferences of client */
	strp = csa_getheaderin(p, "Accept-Language");
	if (strp) {
		if (csa_strcasestr(strp->value, "cz")
		    || csa_strcasestr(strp->value, "cs")
		    || csa_strcasestr(strp->value, "sk"))
		{
			suffix = ".cs";
			outenc = CSTOOLS_ISOLatin2;
		}
		else if (csa_strcasestr(strp->value, "en"))
			suffix = ".en";
	}

	/* test charset preferences of client */
	strp = csa_getheaderin(p, "Accept-Charset");
	if (strp) {
		if (strchr(strp->value, '*')
		    || csa_strcasestr(strp->value, "iso-8859-2"))
		{
			outenc = CSTOOLS_ISOLatin2;
		}
		else {
			cstools_t code;
			const char *name;
			int i;
			for(i=0; i<CSTOOLS_NUMFONTS; i++) {
			    code = cstools_index2code(i);
			    name = cstools_name(code, CSTOOLS_MIMENAME);
			    if (csa_strcasestr(strp->value, name))
			    {
				outenc = code;
				break;
			    }
			}
		}
	}

	/* Netscape 4.0 for Mac sends Accept-Charset header with     */
	/* iso-8859-1 as the only charset, so it's unusable and we   */
	/* have to check the country of remote client anyway */
	if (outenc == CSTOOLS_ISOLatin1) {
		const csa_String *ua = csa_getheaderin(p, "User-Agent");

		if (strncasecmp(ua->value, "Mozilla/4.", 10) == 0
		    && csa_strcasestr(ua->value, "MAC"))
			outenc = CSTOOLS_UNKNOWN;
	}

	/* if we do not know language preferences of client or still    */
	/* don't know charset preferences, use name of remote client    */
	/* to find out if client is from Czech Republic/Slovakia or not */
	if (!suffix || outenc == CSTOOLS_UNKNOWN) {
		const csa_String *remote_host;

		remote_host = csa_getvar(p, "REMOTE_HOST");
		/* if REMOTE_HOST is an IP address, find name */
		if (atoi(remote_host->value)) {
			const char *name;
			name = csa_gethostbyaddr(p->pool_req,
					remote_host->value);
			if (name) {
				csa_setvar(p, "REMOTE_HOST", name, 0);
				remote_host = csa_getvar(p, "REMOTE_HOST");
			}
		}
		
		if (csa_has_suffix(remote_host->value, ".cz,.sk", ','))
		{
			/* every browser in *.cz or *.sk should understand */
			/* iso-8859-2 */
			if (outenc == CSTOOLS_UNKNOWN)
				outenc = CSTOOLS_ISOLatin2;
			if (!suffix)
				suffix = ".cs";
		}
	}

	/* if outenc & suffix is still not set, set it to ASCII & en */
	if (!suffix)	suffix	= ".en";
	if (outenc == CSTOOLS_UNKNOWN) {
		if (strcasecmp(suffix, ".cs") == 0)
			outenc = CSTOOLS_ISOLatin2;
		else
			outenc = CSTOOLS_ASCII;
	}

	if (!redirect_client) {
		/* if we will continue with guessed values (instead of */
		/* redirecting the client), update request structure   */

		p->outcharset = outenc;

		/* force outcharset to be ISO-8859-2 for internall guess */
		/* XXX this needs adjusting once we support non-czech    */
		/* encodings */
		if (p->outcharset != CSTOOLS_ASCII)
			p->outcharset = CSTOOLS_ISOLatin2;

		if (!*p->part.value) {
			csa_fillstring(&p->part, suffix, -1, -1);
			if (strcasecmp(suffix+1, p->dd) == 0)
				CSA_SET(p->flags, CSA_FL_PART_IS_DEF);
			else
				CSA_UNSET(p->flags, CSA_FL_PART_IS_DEF);
		}

		return;
	}

	prefix = csa_getvar(p, "SCRIPT_NAME")->value;

	/* get the part before GUESS or __CHARSET__ out of SCRIPT_NAME and
	 * preserve any suffix the script name might have */
	tocut = csa_strcasestr(prefix, "GUESS");
	if (!tocut) tocut = strstr(prefix, "__CHARSET__");
	if (tocut) {
		prefix = ap_pstrndup(p->pool_tmp, prefix,
			tocut - prefix);
		tocut = strchr(tocut, '.');
		if (tocut) suffix = tocut;
	}

	/* get CSacek name for the guessed encoding */
	charset	= cstools_name(outenc, CSTOOLS_TRUENAME);

	/* create redirect link */
	newurl = (char *) csa_alloca(
		strlen(prefix) + strlen(charset) + strlen(suffix) + 1,
		p->pool_tmp);

	sprintf(newurl, "%s%s%s", prefix, charset, suffix); /* safe */
	newurl = csa_construct_url(p, newurl, NULL);

	csa_setheaderout(p, "Status", "302 Moved Temporarily", 0);
	csa_setheaderout(p, "Location", newurl, CSA_I_COPYVALUE);
}
