#line 2 "loop.c"
/*                                                                      
 *  C-SaCzech
 *
 *  Copyright (c) 1996-2000 Jaromir Dolecek <dolecek@ics.muni.cz>           
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ----
 * Note:
 *   If the conditions of GPL are unacceptable for you, contact
 *   author to adjust something more suitable for your needs.
 */

/* $Id: loop.c,v 1.130 2000/01/24 13:18:07 dolecek Exp $ */

#include "csacek.h" 	
#include "csa_version.h"
#include "csa_cmds.h"

/* local contants */
/* space, which has to be every time free to use in buffer */
#define SAFE_SPACE		32

#define CMD_COMMENT		1  /* cmd has been in HTML comment <!-- .. -->*/
#define CMD_HTMLTAG		2  /* cmd has been HTML tag */
#define CMD_EXTENDED		4  /* cmd not existing in SaCzech */
#define CMD_EXEC_ALWAYS		8  /* execute even if we are in NOCODE block */
#define CMD_NOACCESSCHECK	16 /* execute always without taking care of */
				   /* PART or DOMAIN or CHARSET */
#define CMD_NOESCAPE		32 /* don't recognize '\' as escape char */
				   /* with this cmd */

/* structure used for storing information about each command */
typedef struct {
	const char	*name;          /* name of command */
	int	type;           /* type of command (CSA_CMD0 for example) */
	csa_cmdfunc_t	*func;	/* pointer on function implementing command */
				/* or NULL if explicit rule in cmds.y exists */
	void	*dta;		/* passed to func; usable to give additional */
				/* information to called cmd function */
	int	param_mask;     /* mask of parameters type accepted by cmd */
	int	flags;          /* additional informations */
} csa_cmd_t;

/* local functions */
static csa_cmd_t * x_cmd_lookup __P((const char *, size_t len));
static void x_call_cmd_function __P((csa_params_t *));

/* structure which hold state while processing command */
struct csa_yy {
	csa_arglist_t	 *args;		/* argument for interpreted command */
	csa_cmd_t	 *cmd;		/* pointer to cmd record in x_cmdtab[]*/
	int		  cmd_execd;    /* set after cmd is executed */
	int		  cmd_failed;	/* set if cmd failed to execute */
	int		  state;	/* state of csa_yylex() */
	const char	 *str;		/* text to be interpreted */
	int		  str_len;	/* total number of chars to be read */
	int		  index;	/* index of next char to be read */
	char		 *buf;		/* used while parsing cmd arguments */
#define YY_CURCHAR(YY)	((YY)->str[(YY)->index])
#define YY_VALID(YY)	((YY)->index < (YY)->str_len)
#define YY_REMAINS(YY)	((YY)->str_len - (YY)->index)
	int		 cmd_flags;	/* command flags as passed from
					 * csa_process_body() or any other
					 * caller of csa_run_cmd() */
	int		 igneq;		/* set to true if we encountered */
					/* name=value pair and are parsing */
					/* the value part */
};

/*
 * array of commands supported by CSacek
 * note it's sorted into sub-arrays by the length of command name
 */
#define	CSA_MAX_LENGTH		13
#define CSA_MAX_COMMANDS	6

static csa_cmd_t x_cmdtab[CSA_MAX_LENGTH+1][CSA_MAX_COMMANDS+1] =
{
	{ { NULL, 0, NULL, NULL, 0, 0 } },
	{ /* length 1 */
	  { "A", CSA_CMDX, csa_ChangeURL, (void *)CSA_HREF,
		CSA_P_STRING|CSA_P_EQUATION, CMD_NOESCAPE|CMD_HTMLTAG },
	  { NULL, 0, NULL, NULL, 0, 0 }
	},
	{ { NULL, 0, NULL, NULL, 0, 0 } },
	{ /* length 3 */
	  { "bar", CSA_CMDX, csa_Bar, NULL, CSA_P_STRING|CSA_P_EQUATION,
		CMD_EXTENDED },
	  { "set", CSA_CMDX, csa_Set, NULL, CSA_P_EQUATION,
		CMD_EXEC_ALWAYS|CMD_EXTENDED },
	  { "xml", CSA_CMDX, csa_Xml, NULL, CSA_P_STRING|CSA_P_EQUATION,
		CMD_EXTENDED },
	  { NULL , 0, NULL, NULL, 0, 0 }
	},
	{ /* length 4 */
	  { "BASE", CSA_CMDX, csa_ChangeURL, (void *)CSA_HREF,
		CSA_P_STRING|CSA_P_EQUATION, CMD_NOESCAPE|CMD_HTMLTAG },
	  { "code", CSA_CMD0, csa_Set, (void *)CSA_EXECCMDS, 0,
		CMD_EXEC_ALWAYS|CMD_EXTENDED|CMD_NOACCESSCHECK },
	  { "FORM", CSA_CMDX, csa_ChangeURL, (void *)CSA_ACTION,
		CSA_P_STRING|CSA_P_EQUATION,
		CMD_EXEC_ALWAYS|CMD_NOESCAPE|CMD_HTMLTAG },
	  { "part", CSA_CMDX, csa_DocParts, (void *)CSA_PART,
		CSA_P_STRING|CSA_P_EQUATION, CMD_EXTENDED|CMD_NOACCESSCHECK },
	  { "META", CSA_CMDX, csa_Meta, NULL, CSA_P_EQUATION, CMD_HTMLTAG },
	  { "FONT", CSA_CMDX, csa_Font, NULL, CSA_P_EQUATION, CMD_HTMLTAG },
	  { NULL, 0, NULL, NULL, 0, 0 },
	},
	{ /* length 5 */
	  { "FRAME", CSA_CMDX, csa_ChangeURL, (void *)CSA_SRC,
		CSA_P_STRING|CSA_P_EQUATION, CMD_NOESCAPE|CMD_HTMLTAG },
	  { NULL, 0, NULL, NULL, 0, 0 }
        },
	{ /* length 6 */
 	  { "bardef", CSA_CMDX, csa_BarDef, NULL, CSA_P_STRING|CSA_P_EQUATION,
		CMD_EXTENDED },
	  { "domain", CSA_CMDX, csa_DocParts, (void *)CSA_DOMAIN,
		CSA_P_STRING, CMD_EXTENDED|CMD_NOACCESSCHECK },
	  { "nocode", CSA_CMD0, csa_Set, (void *)CSA_NOEXECCMDS, 0,
		CMD_EXEC_ALWAYS|CMD_EXTENDED },
	  { NULL, 0, NULL, NULL, 0, 0 }
	},
	{ /* length 7 */
	  { "charset", CSA_CMDX, csa_DocParts, (void *)CSA_CHARSET,
		CSA_P_STRING, CMD_EXTENDED|CMD_NOACCESSCHECK},
	  { NULL, 0, NULL, NULL, 0, 0 }
        },
	{ /* length 8 */
	  { "!DOCTYPE", CSA_CMDX, csa_Doctype, NULL, CSA_P_STRING,
		CMD_HTMLTAG },
	  { NULL, 0, NULL, NULL, 0, 0 }
	},
	{ /* length 9 */
	  { "mycharset", CSA_MYCHARSET, NULL,NULL, CSA_P_STRING, CMD_EXTENDED },
	  { NULL, 0, NULL, NULL, 0, 0 }
	},
	{ { NULL, 0, NULL, NULL, 0, 0 } },
	{ { NULL, 0, NULL, NULL, 0, 0 } },
	{ /* length 12 */
	  { /* command csacekServer is obsolete and is provided for
	     * compatibility only */
	    "csacekServer", CSA_CMDX, csa_csacekServers, NULL, CSA_P_STRING,
		CMD_EXEC_ALWAYS|CMD_EXTENDED },
	  { NULL, 0, NULL, NULL, 0, 0 }
	},
	{ /* length 13 */
	  { "csacekServers", CSA_CMDX, csa_csacekServers, NULL, CSA_P_STRING,
		CMD_EXEC_ALWAYS|CMD_EXTENDED },
	  { NULL, 0, NULL, NULL, 0, 0 }
	}
};

/*
 * looks up command ``name'' in table and return associated record
 * or NULL of there is no such command in table
 */
static csa_cmd_t *
x_cmd_lookup (name, len)
  const char *name; /* command name */
  size_t len;	    /* how much chars take for comparison */
{
        int i;
        for (i=0; len<=CSA_MAX_LENGTH && x_cmdtab[len][i].name != NULL; i++)
	{
                if ( CSA_UPPER(x_cmdtab[len][i].name[0]) == CSA_UPPER(name[0])
			&& strncasecmp(x_cmdtab[len][i].name, name, len)==0)
                        return &x_cmdtab[len][i];
	}

        return NULL;
}

/*
 * checks, if cmd can be executed (i.e. current PART & DOMAIN & CHARSET
 * is valid and code is on)
 */
int
csa_cmd_canexecute(p)
  csa_params_t *p;
{
	csa_yy_t *yy = p->yy;
	csa_cmd_t *cmd = yy->cmd;

	return (cmd && !yy->cmd_execd
		&& (CSA_ISSET(p->flags, CSA_VALID_MASK)
		    || CSA_ISSET(cmd->flags, CMD_NOACCESSCHECK)
		   )
		&& (!CSA_ISSET(p->flags, CSA_FL_NOEXECCMDS)
			|| CSA_ISSET(cmd->flags, CMD_EXEC_ALWAYS))
		&& (!CSA_ISSET(cmd->flags, CMD_HTMLTAG)
			|| CSA_ISSET(yy->cmd_flags, CMD_HTMLTAG))
	   );
}

/*
 * sets flag, if cmd failed or not
 */
void
csa_cmd_setfailed(p, failed)
  csa_params_t *p;
  int failed;
{
	p->yy->cmd_failed = failed;
}

/*
 * Gets begin and end tag used in original source for currently
 * processed command.
 * Return 0 if an error occurs, 1 otherwise.
 */
int
csa_yy_gettags(p, startp, endp)
  csa_params_t *p;
  const char **startp, ** endp;
{
	if (CSA_ISSET(p->yy->cmd_flags, CMD_HTMLTAG))
		*startp = "<", *endp = ">";
	else if (CSA_ISSET(p->yy->cmd_flags, CMD_COMMENT))
		*startp = "<!--", *endp = "-->";
	else if (CSA_ISSET(p->yy->cmd_flags, CMD_EXTENDED)) {
		char *startstr, *endstr;
		startstr = (char *)ap_palloc(p->pool_tmp, 3);
		endstr = (char *)ap_palloc(p->pool_tmp, 3);
		sprintf(startstr, "<%c", CSA_CMD_TAG);		/* safe */
		sprintf(endstr, "%c>", CSA_CMD_TAG);		/* safe */
		*startp = startstr, *endp = endstr;
	} else {
		/* the tags are unknown, return error */
		return 0;
	}

	return 1;
}

/* 
 * calls function associated with (previously read) command
 */
static void
x_call_cmd_function(p)
  csa_params_t *p;
{
	int failed=0;

#ifdef CSA_DEBUG
	csa_debug(p->dbg, "x_call_cmd_function: called for command ``%s''",
		(p->yy->cmd) ? p->yy->cmd->name : "<NULL>");
#endif

	if (csa_cmd_canexecute(p) && p->yy->cmd->func)
	{
		failed = p->yy->cmd->func(p, p->yy->cmd->dta);
#ifdef CSA_DEBUG
		csa_debug(p->dbg, "x_call_cmd_function: execution of ``%s'' %s",
			p->yy->cmd->name,
			(failed == 0) ? "succeeded" : "FAILED");
#endif
		p->yy->cmd_execd = 1;
	}
	else if (p->yy->cmd && !p->yy->cmd->func && !p->yy->cmd_execd) {
		/* the command has had eplicit rule in yacc's gramatics */
		p->yy->cmd_execd = 1;
	}
	else failed = 1;

	if (failed) csa_cmd_setfailed(p, failed);
}

/*
 * tries to interpret first ``len'' characters from the string ``s''
 * as a command
 * returns 0 if all is okay
 */
int
csa_run_cmd(p, s, len, cmd_flags)
  csa_params_t *p;
  const char *s;
  size_t  len;
  int cmd_flags;
{
	int retval;

#ifdef CSA_DEBUG
	csa_debug(p->dbg, "csa_run_cmd: str=\"\"\"%.*s\"\"\"" , len, s);
#endif

	p->yy = (csa_yy_t *) ap_pcalloc(p->pool_tmp, sizeof(csa_yy_t));
	p->yy->args = csa_arg_newlist(p->pool_tmp);
	p->yy->str     = s;
	p->yy->str_len = len;
	p->yy->cmd_flags   = cmd_flags;
	
	retval = csa_yyparse(p);

	/* call functions implementing the command, if it has been parsed */
	/* correctly */
	if (retval == 0) {
		x_call_cmd_function(p);
		retval = p->yy->cmd_failed;
	}

#ifdef CSA_DEBUG
	csa_debug(p->dbg, "csa_run_cmd: command %s",
		(retval != 0) ? "failed" : "was succesfull" );
#endif

	return retval;
}

/*
 * returns pointer to list holding arguments passed to command
 */
csa_arglist_t *
csa_yy_getarglist(yy)
  csa_yy_t *yy;
{
	return yy->args;
}

/*
 * returns mask of parameters acceptable to currently executed command
 */
int
csa_yy_getcmdparammask(yy)
  csa_yy_t *yy;
{
	return yy->cmd->param_mask;
}

/*
 * returns name of processed command
 */
const char *
csa_yy_getcmdname(yy)
  csa_yy_t *yy;
{
	return yy->cmd->name;
}

/*
 * recodes given string and adds it to output
 * if copystr is set to not-null, src would never be used for
 * storing result of cstools_recode() (it's would be used otherwise
 * when output encoding is 8bit)
 * if ``buf'' can be null - in that case, big-enough buffer
 * is allocated to hold all necessary data
 */
int
csa_add_recode_output(p, str, len, buf, copystr)
  csa_params_t *p;
  const char *str;
  size_t len;
  csa_String_b *buf;
  int copystr;
{
	size_t dst_len, need_buf_len=0;
	char *output=NULL;

	if (len == 0) return 1;

	if (CSTOOLS_ISUNICODE(p->outcharset)) {
		need_buf_len = 2*len;
	}
	else if (copystr) {
		/* we were explicitely ask to not use input str as output */
		/* buffer as well (it's normally safe to do so, unless str */
		/* is some const string originally */
		need_buf_len = len;
	}
	else {
		/* 8bit encoding - safe to use same string for result of */
		/* recoding, as it would be same or smaller length       */
		output = ap_pstrndup(p->pool_tmp, str, len);
	}
	
	if (need_buf_len || !output || !buf) {
		if (!buf || need_buf_len > buf->maxlen) {
			if (!buf) {
				buf = (csa_String_b *) ap_palloc(p->pool_tmp,
					sizeof(csa_String_b));
				if (!need_buf_len) need_buf_len = len;
			}
			CSA_FILLSTRING(buf, 
				ap_palloc(p->pool_tmp, (int)need_buf_len),
				0, need_buf_len);
		}
		output = buf->value;
	}

	/* handle special windows-1250 characters which do not have */
	/* equivalent in iso-8859-2 - only copyright, registered and */
	/* trade mark sign for now */
	if (p->incharset == CSTOOLS_CP1250 && p->outcharset != CSTOOLS_CP1250)
	{
		char outbuf[2*sizeof("&trade;") + 1];
		const char *substs[3][2] = {
			{ "&copy;",	"(c)" },
			{ "&reg;",	"(R)" },
			{ "&trade;",	"(TM)" } };
		const char *subst;
		size_t sindex, lastindex=0, offset;

		offset = CSA_ISSET(p->flags, CSA_FL_ISHTML) ? 0 : 1;
		
		for(sindex=0; sindex < len; sindex++) {
		    if (CSA_CHCMP(str[sindex], 0xa9)) {
			/* copyright sign */
			subst = substs[0][offset];
		    }
		    else if (CSA_CHCMP(str[sindex], 0xae)) {
			/* registered sign */
			subst = substs[1][offset];
		    }
		    else if (CSA_CHCMP(str[sindex], 0x99)) {
			/* trade mark sign */
			subst = substs[2][offset];
		    }
		    else {
			/* not a special char, continue */
			continue;
		    }
		    dst_len = cstools_recode(&(p->mp), &(str[lastindex]),
			output, (int) (sindex - lastindex));
		    csa_add_output(p, output, dst_len, 0);
		    dst_len = cstools_recode(&(p->mp), subst, outbuf,
			(int) strlen(subst));
		    csa_add_output(p, outbuf, dst_len, 0);
		    lastindex = sindex + 1;
		}
		if (lastindex >= len) return 0; /* nothing more to do */
		if (lastindex > 0) {
			str = &(str[lastindex]);
			len -= lastindex;
		}
	}
	
	dst_len = cstools_recode(&(p->mp), str, output, (int) len);
	csa_add_output(p, output, dst_len, 0);

	return 0;
}

/*
 * makes supported substitutes and recoding on the supplied text
 * and adds it to output; ``len'' chars from input are processed
 * if ``add_all'' is non-zero, all output is flushed to output;
 * otherwise code checks if the text doesn't end by fragment of
 * __FOO__ string and if so, the fragment is not added to output
 * and it's copiend on the beginning of the ``text'' string
 */
int
csa_add_subs_output(p, text, out_len, add_all)
  csa_params_t *p;
  csa_String_b *text;
  size_t out_len;
  int add_all;
{
	int num_processed=out_len;

#ifdef CSA_DEBUG
	csa_debug(p->dbg, "csa_add_subs_output: called");
#endif

	if (CSA_ISSET(p->flags, CSA_VALID_MASK))
	{
	    csa_String_b buf;
	    const csa_String *subs;
	    const char *where, *behind, *laststr;
	    char overwritten_char;
	    size_t last_chunk_len;

	    buf.value = NULL;
	    buf.maxlen = buf.len = 0;
	    if (!CSTOOLS_ISUNICODE(p->outcharset))
		buf.value = text->value;

	    overwritten_char = text->value[out_len];
	    text->value[out_len] = 0;
	    
	    laststr = text->value;
	    while(csa_find_subs(p, laststr, &where, &behind, &subs) ) {

		/* put all the text up to "__" to output */
		csa_add_recode_output(p, laststr, (size_t) (where - laststr),
			&buf, 0);

		/* add to output apropriate substitute for __FOO__ */
		csa_add_recode_output(p, subs->value, subs->len, &buf, 1);

		laststr = behind;
	    } /* while() */

	    last_chunk_len = strlen(laststr);
	    if (!add_all && last_chunk_len) {
                int checklen = (last_chunk_len > 20) ? 20 : last_chunk_len;
		char *temp = strchr(laststr, '\0'); /* find end of string */
		temp--;
		for(;temp > laststr && checklen > 0; checklen--, temp--)
			if (*temp == '_') {
				if (*(temp-1) == '_' || *(temp+1) == '\0') {
                        		int len = temp - laststr;
                        		num_processed -= (last_chunk_len - len);
                        		last_chunk_len = len;
				}
				break;
			}
	    }

	    csa_add_recode_output(p, laststr, last_chunk_len, &buf, 0);
		
	    text->value[out_len] = overwritten_char;
	}
#ifdef CSA_DEBUG
	else 
	    csa_debug(p->dbg, "csa_add_subs_output: part output flags invalid \
(%d), string discarded",
			p->flags & CSA_VALID_MASK);
#endif

	text->len -= num_processed;
	if (text->len) {
		strncpy(text->value, &text->value[num_processed], text->len);
	}
	text->value[text->len] = '\0';

	return 0;
}

/* 
 * reads tokens and returns their identification to cmds.c:yyparse()
 */
int 
csa_yylex (p, parp)
  csa_params_t *p;
  void *parp;
{
	csa_cmd_t *cmditem;
	size_t sindex;
	YYSTYPE *lvalp=parp;
	int retval=0;
	csa_yy_t *yy = p->yy;

#ifdef CSA_DEBUG
	csa_debug(p->dbg, "yylex: state %d, bytes to read: %d", yy->state,
		yy->str_len - yy->index);
#endif

	/* ignore leading white space */
	while(YY_VALID(yy) && isspace((int)YY_CURCHAR(yy)) )
		yy->index++;

	/* if nothing more remains in the buffer, exit now */
	if (!YY_VALID(yy)) {
		return 0;
	}

	switch(yy->state) {
	case 0: /* start state before we find out name of command */
		if (yy->cmd != NULL) {
			x_call_cmd_function(p);
			yy->cmd = NULL;
		}

		sindex = yy->index;
		while(YY_VALID(yy)
			&& !isspace((int)YY_CURCHAR(yy))
			&& YY_CURCHAR(yy) != '='
			&& YY_CURCHAR(yy) != ';' ) yy->index++;
		cmditem = x_cmd_lookup(&(yy->str[sindex]),
					yy->index - sindex);

		/* if there was found command slot, fill in */
		/* p->yy with appropriate values and update */
		/* retval accordingly */
		if (cmditem) {
			lvalp->res.s = cmditem->name;
			yy->cmd = cmditem;			
			yy->state = 1;
			yy->cmd_execd = 0;
			retval = cmditem->type;
		}
		else {
			/* nothing suitable found, return end-of-input */
			retval = 0;
		}

		break;

	case 1: /* reading arguments */
		if (YY_CURCHAR(yy) == '=' && !yy->igneq)
		{
			lvalp->res.s = "=";
			yy->index++;
			yy->igneq = 1;
			retval = '=';
		}
		else if (YY_CURCHAR(yy) == ';')
		{
			lvalp->res.s = ";";
			yy->index++;
			yy->state = 0;
			yy->igneq = 0;/* just in case this follows eq sign */
			retval = ';';
		}
		else {
			char *outp, quote = '\0', curchar, useq = '\0';
			int escapechar=0;

			if (!yy->buf) {
				yy->buf = (char *)
					ap_palloc(p->pool_tmp,
					YY_REMAINS(yy) + 1);
			}
			outp = yy->buf;
			for(;YY_VALID(yy); yy->index++)
			{
				curchar = YY_CURCHAR(yy);

				if (escapechar) {
					/* this character has been escaped, */
					/* do not check if it's special */
					escapechar = 0;
					goto csa_yylex_addchar;
				}

				/* check quotes first */
				if (curchar == '\'' || curchar == '"') {
					if (!useq)
						useq = curchar;

					if (!quote) {
						quote = curchar;
						continue;
					}
					else if (quote == curchar) {
						quote = '\0';
						continue;
					}
				}

				/* break processing if the char is not */
				/* quoted and it's ';' or space or '=' */
				/* which is not to be ignored */
				if (!quote && 
					(curchar == ';' || isspace((int)curchar)
					 || (curchar == '=' && !yy->igneq)) )
				{
					break;
				}

				/* special case \\ (escape char);    */
				/* if we encounter it and escaping   */
				/* is not disallowed for the char we */
				/* are processing, schedule for next */
				/* char to be not treated specially  */
				if (curchar == '\\'
				    && !(yy->cmd->flags & CMD_NOESCAPE) )
				{
					escapechar = 1;
					continue;
				}

			    csa_yylex_addchar:
				/* if we get here, we really want to */
				/* add the char to output string     */
				*(outp++) = curchar;
			} /* for */

			*outp = '\0'; /* end the string */
			lvalp->res.s = yy->buf;
			lvalp->res.opaque = useq;
			yy->igneq = 0; /* clear igneq flag */

			retval = CSA_STRING;
		} /* else branch for state == 1 */
	}
			
	return retval;
}
	
/*
 * reads input data, processes them and prepares them for output
 * returns 0 in case of normal termination; returns CSA_DONE, if all
 * output was sent to client; calling code should NOT call csa_output()
 * then
 * we have to be able to re-enter this function in case response
 * data are not yet available, so count with it
 */
int 
csa_process_body(p)
  csa_params_t *p;
{
  csa_String_b *buf;
  char quote, *temp, *cmd;
  int cmd_index, cmd_flags, leave;
  size_t output_len, len, sindex;
  int escaped, add_all=0, need_read_more=1, free_tmp_pool=1;

#ifdef CSA_DEBUG
   csa_debug(p->dbg, "process_body: begin");
#endif

  /* init process buffer */
  if (!p->body_buf) {
	p->body_buf = (csa_String_b *) ap_palloc(p->pool_req,
					sizeof(csa_String_b));
	CSA_FILLSTRING(p->body_buf, (char *) ap_palloc(p->pool_req, 8192),
		0, 8192);
  }
  buf = p->body_buf;

#ifdef CSA_DEBUG
	csa_debug(p->dbg, "process_body: about to enter process loop");
#endif

  for(len=1; len || buf->len; ) {

#ifdef CSA_DEBUG
	csa_debug(p->dbg, "process_body: next cycle");
#endif

	/* free space allocated in temporary pool */
	if (free_tmp_pool) ap_clear_pool(p->pool_tmp);
	else free_tmp_pool = 1;

	/* read a chunk, if needed */
	if (need_read_more || buf->len == 0) {
		len = csa_read_response(p, &buf->value[buf->len],
				buf->maxlen - buf->len - 1);
		if (len == 0)
			break; /* no more data available, exit */
		buf->len += len;
		buf->value[buf->len] = '\0'; /* mark an end */
		need_read_more = 0;
	}

	/* send data directly to output if we won't change it */
	if (!CSA_ISSET(p->flags, CSA_FL_CONVERT)) {
		csa_add_output(p, buf->value, buf->len, 0);
		buf->len = 0;
		continue;
	}

	cmd_index = 0;
	cmd_flags = 0;
	temp = buf->value;
	for(;;) {
		temp = strchr(temp, '<');
		if (!temp) break;

		if ((buf->len - (temp - buf->value)) <= 5) {
			need_read_more = 1;
			break;
		}

		cmd = temp + 1;

		if (cmd[0] == CSA_CMD_TAG) {
			cmd_index = 2;
			CSA_SET(cmd_flags, CMD_EXTENDED);
			break;
		}
		else if (strncmp(cmd, "!--", 3) == 0) {
			cmd_index = 4;
			CSA_SET(cmd_flags, CMD_COMMENT);
			break;
		}
		else if (strncasecmp(cmd, "META", 4) == 0)
		{
			cmd_index = 1;
			CSA_SET(cmd_flags, CMD_HTMLTAG);
			break;
		}
		else if (CSA_ISSET(p->flags, CSA_FL_ISHTML) &&
		    ((CSA_ISSET(p->flags, CSA_FL_CHANGEURL)
			&& (CSA_UPPER(cmd[0]) == 'A'
				|| strncasecmp(cmd, "FORM", 4) == 0
				|| strncasecmp(cmd, "BASE", 4) == 0
				|| strncasecmp(cmd, "FRAME", 5) == 0))
		    || strncasecmp(cmd, "!DOCTYPE", 8) == 0
		    || strncasecmp(cmd, "FONT", 4) == 0) )
		{
			cmd_index = 1;
			CSA_SET(cmd_flags, CMD_HTMLTAG);
			break;
		}

		temp++;
	}
				
	add_all = 0;
	if (temp) {
		if (!need_read_more) add_all = 1;
		output_len = temp - buf->value;
	}
	else
		output_len = buf->len;
		
	/* add already read data into output */
	if (output_len) csa_add_subs_output(p, buf, output_len, add_all);

	if (!cmd_index) {
		free_tmp_pool = 0;
		continue; /* no command found, continue reading input */
	}

	sindex = cmd_index;
	leave = escaped = 0;
	quote = '\0';
	for(;;sindex++) {
		size_t tail_len;

		/* allocate bigger buffer if we need one */
		if (sindex + SAFE_SPACE >= buf->maxlen) {
			buf->maxlen *= 2;
			temp = (char *) ap_palloc(p->pool_req,(int)buf->maxlen);
			strcpy(temp, buf->value); /* safe */
			buf->value = temp;
		}

		/* read more data, if needed */
		if (sindex >= buf->len) {
			len = csa_read_response(p, &buf->value[buf->len],
					buf->maxlen - buf->len - 1);

			/* end of input encountered before command */
			/* has been recognized */
			if (!len) {
				need_read_more = 1;
				break;
			}

			buf->len += len;
			buf->value[buf->len] = '\0';
		}

		/* ignore char if it's escaped */
		if (escaped) {
			escaped = 0;
			continue;
		}

		switch(buf->value[sindex]) {
			case '\\':
				/* backslash in quotes escapes next char */
				/* any time we get here, ``escaped'' is 0 */
				if (quote != '\'') escaped = 1;
				break;

			case '\'':
			case '"':
				if (quote == '\0') quote = buf->value[sindex];
				else if (quote==buf->value[sindex]) quote ='\0';
				break;

			case '>':
				if (quote && !CSA_ISSET(cmd_flags, CMD_COMMENT))
					break;

				if (CSA_ISSET(cmd_flags, CMD_COMMENT)
					&& buf->value[sindex-1] == '-'
					&& buf->value[sindex-2] == '-')
						tail_len = 3;
				else if (CSA_ISSET(cmd_flags, CMD_EXTENDED)
					&& buf->value[sindex-1] == CSA_CMD_TAG)
						tail_len = 2;
				else if (CSA_ISSET(cmd_flags, CMD_HTMLTAG))
						tail_len = 1;
				else /* just a random '>' */
						break;

				if (csa_run_cmd(p, &buf->value[cmd_index],
					sindex - cmd_index - tail_len+1,
					cmd_flags) == 0)
				{
					/* succesful completition */
					strcpy(buf->value,
						&buf->value[sindex+1]);
					buf->len -= sindex+1;
					buf->value[buf->len] = '\0';
				}
				else csa_add_subs_output(p, buf,
					(size_t)sindex+1, 1);

				leave = 1;
				break;
		} /* switch */

		if (leave) break; /* end parsing */
	} /* cmd parsing for loop */

	/* we would need read some more data, but it's not available, so quit */
	if (need_read_more) break;

  } /* for */

  /* Note: remains in buffer are handled by x_finish_body() - we can't */
  /* flush it now as csa_process_body() can be called again when there are */
  /* more data available */

  /* free tmp pool */
  ap_clear_pool(p->pool_tmp);

#ifdef CSA_DEBUG
   csa_debug(p->dbg, "process_body: end");
#endif
 
   return 0;
} /* csa_process_body */
