#line 2 "servlist.c"
/*                                                                      
 *  C-SaCzech
 *
 *  Copyright (c) 1996-2000 Jaromir Dolecek <dolecek@ics.muni.cz>           
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ----
 * Note:
 *   If the conditions of GPL are unacceptable for you, contact
 *   author to adjust something more suitable for your needs.
 */

/* $Id: servlist.c,v 1.5 2000/01/24 13:18:09 dolecek Exp $ */

#include "csacek.h"

typedef struct {
	const char	*name;
	size_t		 sn_len;
	int	 port;
	int	 chksum;
} csa_server_t;

struct csa_slist {
	size_t max_len[4];
	size_t len[4];
	csa_server_t *servers[4];
	const struct csa_slist *other;
	struct csa_slist * const *otherp;
};

#define NEW_SLIST(wpool) (csa_slist_t *)ap_pcalloc(wpool, sizeof(csa_slist_t))
#define COMPUTE_CHKSUM(name, chksum, len, i, hash) \
	chksum = 0;				\
	len = strlen(name);			\
	for(i=0; i< len; i++) { 		\
		chksum += CSA_UPPER(name[i]);	\
	}					\
	hash = chksum % 4;

/*
 * Following if condition might be a bit hard to parse, don't look
 * if you are easy to be scared :)
 *
 * The comparing algorith is as follows:
 * 1. if the feature is higher in "new" item, it is greater and
 *    condition is TRUE
 * 2. if the feature is equal in both the "new" item and
 *    listitem, take "lower" level feature and goto 1.
 * 3. if the feature is lower in the inserted item, listitem
 *    is greater and condition is FALSE
 */
#define ITEM_IS_EQ_OR_GREATER(chksum, port, len, name, listitem) \
	(chksum > listitem->chksum ||					\
	    (chksum == listitem->chksum && (port > listitem->port ||	\
		(port == listitem->port && (len > listitem->sn_len ||	\
		    (len == listitem->sn_len				\
			&& strcasecmp(name, listitem->name) >= 0))))))	\

/*
 * add server + port to list of CSacek servers
 */
int
csa_addserver(wpool, listp, name, port)
  struct pool *wpool;
  csa_slist_t **listp;
  const char *name;
  int port;
{
	size_t len, i, old_len, new_len, idx, lim;
	int chksum, hash;
	csa_server_t *new_sarr, *s, *p, *base;
	csa_slist_t *list;

	if (!name || !*name) {
		/* empty server name, no reason why we should add the recod */
		/* to list */
		return 1;
	}

	if (!*listp)
		*listp = NEW_SLIST(wpool);
	list = *listp;

	COMPUTE_CHKSUM(name, chksum, len, i, hash);

	/*
	 * Allocate more space if the hash list is full already.
	 */
	if (list->max_len[hash] == list->len[hash]) {
		old_len = list->len[hash];
		new_len = (!old_len) ? 8 : (old_len<<1);
		new_sarr = (csa_server_t *)ap_palloc(wpool,
				(int)(new_len * sizeof(csa_server_t)));
		if (old_len > 0) {
			memcpy(new_sarr, list->servers[hash],
				old_len * sizeof(csa_server_t));
		}
		list->servers[hash] = new_sarr;
		list->max_len[hash] = new_len;
	}

	/*
	 * Arrange things so that the entries are sorted by checksum, port,
	 * length of name and name. Use the fact that previous items are
	 * sorted and find the spot where should the new entry be put
	 * with derivate of binary search. Note we don't check whether
	 * the same item is on the list already - such thing should
	 * not happen very often and it's not worth the wasted CPU cycles.
	 */
	base = &list->servers[hash][0]; 
	lim = list->len[hash];
	idx = 0;
	for(; lim != 0; lim >>= 1) {
		p = &base[lim >> 1];
		if (ITEM_IS_EQ_OR_GREATER(chksum, port, len, name, p)) {
			idx += (lim >> 1) + 1;
			base = &p[1];
			lim--;
			continue;
		}
	}

	/*
	 * If necessary, shift data to create place for the new entry.
	 */
	if (idx < list->len[hash]) {
		p = &list->servers[hash][idx];
		memmove(&p[1], p, (list->len[hash] - idx)*sizeof(csa_server_t));
	}

	/*
	 * Fill in the spot.
	 */
	s = &list->servers[hash][idx];
	s->name   = name;
	s->sn_len = len;
	s->port   = port;
	s->chksum = chksum;

	list->len[hash]++;

	return 0;
}

/*
 * Embed list in other list.
 */
void
csa_append_list(wpool, tlistp, slist)
  struct pool *wpool;
  csa_slist_t **tlistp;
  const csa_slist_t *slist;
{
	/* If the list is empty, try it's predecessor and so on */
	for(; slist; slist = slist->other) {
		if (slist->len[0] || slist->len[1] || slist->len[2]
		    || slist->len[3])
			break;
	}
	if (!slist) return;

	if (!*tlistp) *tlistp = NEW_SLIST(wpool);
	(*tlistp)->other = slist;
}

/*
 * Embed list in other list. No check for predecessor is made, so
 * slistp doesn't need to point to currently valid data.
 */
void
csa_append_listp(wpool, tlistp, slistp)
  struct pool *wpool;
  csa_slist_t **tlistp, **slistp;
{
	if (!slistp) return;

	if (!*tlistp) *tlistp = NEW_SLIST(wpool);
	(*tlistp)->otherp = slistp;
}
/*
 * Checks if the server is on list of CSacek servers.
 */
int
csa_is_csacek_server(slist, name, port)
	const csa_slist_t *slist;
	const char *name;
	int port;
{
	const csa_slist_t *list;
	const csa_server_t *base, *p;
	int chksum, hash;
	size_t i, len, lim;

	if (!name) return 1;
	if (!slist)  return 0;

	COMPUTE_CHKSUM(name, chksum, len, i, hash);

	list = slist;
   loop:
	/*
	 * Search the list. Alhorith taken from implementation of
	 * NetBSD bsearch(3).
	 */
	base = &list->servers[hash][0];
	lim = list->len[hash];
	for(; lim != 0; lim >>= 1) {
		p = &base[lim >> 1];
	 	if (chksum == p->chksum && (!p->port || port == p->port)
			&& len == p->sn_len
			&& strcasecmp(name, p->name) == 0)
		{
			return (1);
		}

		if (ITEM_IS_EQ_OR_GREATER(chksum, port, len, name, p)) {
			base = &p[1];
			lim--;
		}
	}
	list = (list->other) ? list->other
			: ((list->otherp) ? *list->otherp : NULL);
	if (list)
		goto loop;

	return (0);
} 

#ifdef CSA_WANT_INFO
/*
 * Writes servers from list of CSacek servers into output.
 */
void
csa_write_servers(p)
  csa_params_t *p;
{
	const csa_slist_t *slist = p->csacek_servers;
	csa_server_t *serv;
	char buf[2048];
	size_t len, listlen, hash=0, i;

	if (!slist) return;

    loop:
	serv = slist->servers[hash];
	listlen = slist->len[hash];
	for(i=0; i < listlen; i++) {
	    if (serv->port)  {
		len = sprintf(buf, "<TT>%s:%d</TT>, ", serv->name,
				serv->port);
	    } else
		len = sprintf(buf, "<TT>%s</TT>, ", serv->name);
	    csa_add_output(p, buf, len, 0);
	}
	hash++;
	if (hash == 4) {
		slist = slist->other;
		hash = 0;
	}
	if (slist)
		goto loop;
}
#endif /* CSA_WANT_INFO */
