#line 2 "string.c"
/*                                                                      
 *  C-SaCzech
 *
 *  Copyright (c) 1996-2000 Jaromir Dolecek <dolecek@ics.muni.cz>           
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ----
 * Note:
 *   If the conditions of GPL are unacceptable for you, contact
 *   author to adjust something more suitable for your needs.
 */

/* $Id: string.c,v 1.51 2000/01/24 13:18:11 dolecek Exp $ */

/* CSacek functions which work with strings */

#include "csacek.h"

/* 
 * tries to find ``search'' in ``str''; the comparison is case insensitive
 * note this routine works for ASCII only
 */                   
const char *
csa_strcasestr( str, search )
  const char *str, *search;
{
	const char *i,*j;

	if (!*search) return str;

   	for(; *str; str++) {
	   if (CSA_UPPER(*str) == CSA_UPPER(*search)) {
		i = str+1;
		j = search+1;
		for(; *i && *j && CSA_UPPER(*i) == CSA_UPPER(*j); i++, j++);
		if (*j == '\0') return str;
	   }
	}

	return NULL;
}

/*
 * fills csa_String with values given; if ``len'' is lower than 0,
 * strlen(val) is supstutited instead of it; if ``maxlen'' is lower
 * than zero, ``len'' is subtituted instead of it
 */
void
csa_fillstring(str, val, len, maxlen)
  csa_String *str;
  const char   *val;
  int len, maxlen;
{
	str->value = val;
	str->len = (len >= 0) ? len : strlen(val);
	str->maxlen = (maxlen >= 0) ? maxlen : str->len;
}

/* creates & fills in new csa_String struct */
csa_String *
csa_createstring(wpool, val)
  struct pool *wpool;
  const char *val;
{
	csa_String *newstr;
	newstr = (csa_String *) ap_palloc(wpool, sizeof(csa_String));
	csa_fillstring(newstr, val, -1, -1);
	return newstr;
}

/*
 * finds out, if ``str'' ends with one of items in ``list'';
 * items in ``list'' are separated by ``sep''
 * returns pointer at suffix found in  string ``str'' or NULL if no suffix
 * was found 
 */
const char * 
csa_has_suffix(str, list, sep_int)
  const char *str, *list;
  int sep_int;
{
   const char *b,  *suffix_pos=NULL;
   char sep=(char)sep_int;
   int str_len = 0, len = 0;

   str_len = strlen(str);
   while( *list )
   {
      b = strchr(list, sep);
      if (b == NULL) b = strchr(list,0); /* last item */
      len = b - list;
      if ( len <= str_len && len > 0 
		/* LINTED */
		&& strncmp(list, &str[str_len-len], len) == 0 )
      {
	suffix_pos = &str[str_len-len];
	break;
      }
      if (*b) list = ++b; /* shift list behind separator */
      else list = b;
   }

   return suffix_pos;
}                                             

/*
 * this finds first __FOO__ string and returns position where it was found,
 * pposition of text behind it and csa_String structure with what it
 * should be substituted by
 * return value if non-zero if some __FOO__ has been found, 0 otherwise
 */
int
csa_find_subs(p, str, where, behind, subs)
  csa_params_t *p;
  const char *str;
  const char **where, **behind;
  const csa_String **subs;
{
	const char *temp;
	int add_len=0;

	temp = str;
	for(;;) {
		temp = strstr(temp, "__");
		if (!temp) break;
	
		if (strncasecmp(temp, "__CHARSET__", 11) == 0) {
			add_len = 11;
			*subs	= &p->charset;
		}
		else if (strncasecmp(temp, "__PART__", 8) == 0) {
			add_len = 8;
			*subs	= &p->part;
		}
		else if (strncasecmp(temp, "__LAMPACHARSET__", 16) == 0) {
			add_len = 16;
			*subs	= &p->lampacharset;
		}
		else {
			temp++;
			continue; /* nothing supported */
		}

		*where	= temp;
		*behind	= temp + add_len;
		return 1;
	}

	return 0; /* nothing supported found */
}

/*
 * parses string and substitutes all occurencies of supported __FOO__
 * strings by appropriate values; returns new strings with all
 * __FOO__ strings substituted or original string if no __FOO__ string
 * has been found in it
 * space for new string is allocated in temporary pool, so it will be
 * reused as soon as we reach main loop in loop.c again
 */
const char *
csa_subs_string(p, str)
  csa_params_t *p;
  const char *str;
{
	const char *where, *behind;
	char *temp=NULL;
	const csa_String *subs;
	size_t buff_len=0, need_len, copy_len;

	while((csa_find_subs(p, str, &where, &behind, &subs))) {
		copy_len = where - str;
		need_len = copy_len + subs->len + strlen(behind) + 1;
		if (!temp || buff_len < need_len) {
			temp = (char *)csa_alloca(need_len, p->pool_tmp);
			buff_len = need_len;
			if (copy_len) strncpy(temp, str, copy_len);
		}
		strncpy(&temp[copy_len], subs->value, subs->len);
		strcpy(&temp[copy_len + subs->len], behind);
		str = temp;
	}

	return (temp) ? ap_pstrdup(p->pool_tmp, temp) : str;
}

/*
 * It checks whether strings are non-null before comparing them. 
 * Returns: 0 - same, 1 - different or any is null
 */
int
csa_n_strcmp(s1, s2)
  const char *s1, *s2;
{
    return !(s1 && s2 && strcmp(s1, s2) == 0);
}
